@@ -1,8 +1,12 @@
 """Pyinstaller hooks for the main module."""
-from PyInstaller.utils.hooks import get_module_file_attribute
 import os
 
-srctools_loc = os.path.dirname(get_module_file_attribute('srctools'))
+from PyInstaller.utils.hooks import get_module_file_attribute  # type: ignore
+
+
+srctools_init = get_module_file_attribute('srctools')
+assert srctools_init is not None
+srctools_loc = os.path.dirname(srctools_init)
 
 datas = [
     # Add our FGD database.
