@@ -1,14 +1,27 @@
 """Implements support for collapsing instances."""
+from typing import (
+    Any, Container, Dict, Iterable, List, MutableMapping, Optional, Set, Tuple, Union,
+)
 from enum import Enum
 from pathlib import Path
-from typing import Union, Tuple, Dict, Set, Iterable, Container, List
-from srctools import Matrix, Vec, Angle, conv_float, Property
-from srctools.vmf import Entity, EntityFixup, FixupValue, VMF, Output, VisGroup
-from srctools.fgd import ValueTypes, FGD, EntityDef, EntityTypes
-from srctools.filesys import FileSystemChain, RawFileSystem, FileSystem
+import warnings
+
+import attrs
+
+from srctools.fgd import FGD, EntityDef, EntityTypes, ValueTypes
+from srctools.filesys import FileSystem, FileSystemChain, RawFileSystem
+from srctools.keyvalues import Keyvalues
+from srctools.math import Angle, Matrix, Vec, format_float
+from srctools.vmf import VMF, Entity, EntityFixup, FixupValue, Output, VisGroup
+import srctools
 import srctools.logger
 
 
+__all__ = [
+    'FixupStyle',
+    'Instance', 'Manifest', 'Param', 'InstanceFile',
+    'get_inst_locs', 'collapse_all', 'collapse_one',
+]
 LOGGER = srctools.logger.get_logger(__name__)
 # Hidden variable to track the number of recursions.
 RECUR_COUNT_ATTR = '_inst_recur_count'
@@ -18,21 +31,35 @@ _UNKNOWN_KV: Set[Tuple[str, str]] = set()
 
 class FixupStyle(Enum):
     """The kind of fixup style to use."""
-    PREFIX = 0
-    SUFFIX = 1
-    NONE = 2
+    PREFIX = 0  #: Entities will be named :samp:`{inst_name}-{ent_name}`.
+    SUFFIX = 1  #: Entities will be named :samp:`{ent_name}-{inst_name}`.
+    NONE = 2  #: Entities will remain named :samp:`{ent_name}`.
 
 
 class Instance:
     """Represents an instance with all the values required to collapse it."""
+    name: str
+    filename: str
+    pos: Vec
+    orient: Matrix
+    fixup_type: FixupStyle
+    fixup: EntityFixup
+    recur_count: int
+    outputs: List[Output]
+
+    ent_ids: Dict[int, int]
+    face_ids: Dict[int, int]
+    brush_ids: Dict[int, int]
+    node_ids: Dict[int, int]
+    visgroup_ids: Dict[int, int]
     def __init__(
         self,
         name: str,
         filename: str,
         pos: Vec, orient: Matrix,
         fixup_type: FixupStyle,
-        outputs: Iterable[Output]=(),
-        fixup: Iterable[FixupValue]=(),
+        outputs: Iterable[Output] = (),
+        fixup: Iterable[FixupValue] = (),
     ) -> None:
         self.name = name
         self.filename = filename
@@ -42,11 +69,11 @@ class Instance:
         self.fixup = EntityFixup(fixup)
         self.outputs = list(outputs)
         # After collapsing, this is the original -> new ID mapping.
-        self.ent_ids: Dict[int, int] = {}
-        self.face_ids: Dict[int, int] = {}
-        self.brush_ids: Dict[int, int] = {}
-        self.node_ids: Dict[int, int] = {}
-        self.visgroup_ids: Dict[int, int] = {}
+        self.ent_ids = {}
+        self.face_ids = {}
+        self.brush_ids = {}
+        self.node_ids = {}
+        self.visgroup_ids = {}
         # Keep track of recursive instances to handle loops.
         self.recur_count = 0
 
@@ -67,7 +94,7 @@ class Instance:
             name,
             filename,
             Vec.from_str(ent['origin']),
-            Matrix.from_angle(Angle.from_str(ent['angles'])),
+            Matrix.from_angstr(ent['angles']),
             fixup_style,
             ent.outputs,
             ent.fixup.copy_values(),
@@ -97,10 +124,12 @@ class Instance:
     ) -> str:
         """Transform this keyvalue to the new instance's location and name.
 
-        - classnames is a set of known entity classnames, used to avoid renaming
-        those.
+        :param vmf: The rest of the map.
+        :param classnames: This is a set of known entity classnames, used to avoid renaming those.
+        :param type: The kind of FGD value.
+        :param value: The value of the key.
         """
-        # All three are absolute positions.
+        # All three of these types are absolute positions.
         if type is ValueTypes.VEC or type is ValueTypes.VEC_ORIGIN or type is ValueTypes.VEC_LINE:
             return str(Vec.from_str(value) @ self.orient + self.pos)
         elif type is ValueTypes.ANGLES:
@@ -108,7 +137,8 @@ class Instance:
         elif type.is_ent_name:  # Target destination etc.
             return self.fixup_name(value)
         elif type is ValueTypes.TARG_DEST_CLASS:
-            # Target destination, but also classnames which we don't want to change.
+            # Accepts both entity names and classnames - if this is a classname, pass it through
+            # unaltered.
             if value.casefold() not in classnames:
                 return self.fixup_name(value)
         elif type is ValueTypes.EXT_VEC_DIRECTION:
@@ -125,8 +155,6 @@ class Instance:
                     sides.append(str(new_side))
             sides.sort()
             return ' '.join(sides)
-        elif type is ValueTypes.VEC_AXIS:
-            value = str(Vec.from_str(value) @ self.orient)
         elif type is ValueTypes.TARG_NODE_SOURCE or type is ValueTypes.TARG_NODE_DEST:
             # For each old ID always create a new ID.
             try:
@@ -140,9 +168,9 @@ class Instance:
                 value = str(new_id)
         elif type is ValueTypes.VEC_AXIS:
             # Two positions seperated by commas.
-            first, second = value.split(',')
-            first = Vec.from_str(first) @ self.orient + self.pos
-            second = Vec.from_str(second) @ self.orient + self.pos
+            first_str, second_str = value.split(',')
+            first = Vec.from_str(first_str) @ self.orient + self.pos
+            second = Vec.from_str(second_str) @ self.orient + self.pos
             value = f'{first}, {second}'
         # All others = no change required.
         return value
@@ -166,7 +194,7 @@ class Manifest(Instance):
         self.is_toplevel = is_toplevel
 
     @classmethod
-    def parse(cls, tree: Property) -> List['Manifest']:
+    def parse(cls, tree: Keyvalues) -> List['Manifest']:
         """Parse a VMM file."""
         return [
             cls(
@@ -177,32 +205,32 @@ class Manifest(Instance):
         ]
 
 
+@attrs.define
 class Param:
     """Configuration for a specific fixup variable."""
-    def __init__(
-        self,
-        name: str,
-        type: ValueTypes=ValueTypes.STRING,
-        default: str='',
-    ) -> None:
-        self.name = name
-        self.type = type
-        self.default = default
+    name: str
+    type: ValueTypes = ValueTypes.STRING
+    default: str = ''
 
 
 class InstanceFile:
     """Represents an instance VMF which has been parsed."""
+    vmf: VMF
+    params: Dict[str, Param]
+    #: Inputs into the instance. The key is the parts of the instance:name;input string.
+    proxy_inputs: Dict[Tuple[str, str], Output]
+    #: Outputs out of the instance. The key is the parts of the instance:name;output string.
+    #: The value is the ID of the entity to add the output to.
+    proxy_outputs: Dict[Tuple[str, str], Tuple[int, Output]]
+
+    #: If instructed to add in a proxy later, this is the local pos to place it.
+    proxy_pos: Vec
+
     def __init__(self, vmf: VMF) -> None:
         self.vmf = vmf
-        self.params: Dict[str, Param] = {}
-        # Inputs into the instance. The key is the parts of the instance:name;input string.
-        self.proxy_inputs: Dict[Tuple[str, str], Output] = {}
-        # Outputs out of the instance. The key is the parts of the instance:name;output string.
-        # The value is the ID of the entity to add the output to.
-        self.proxy_outputs: Dict[Tuple[str, str], Tuple[int, Output]] = {}
-
-        # If instructed to add in a proxy later, this is the local pos to place
-        # it.
+        self.params = {}
+        self.proxy_inputs = {}
+        self.proxy_outputs = {}
         self.proxy_pos = Vec()
 
         self.parse()
@@ -211,7 +239,7 @@ class InstanceFile:
         """Parse func_instance_params and io_proxies in the map."""
         for params_ent in self.vmf.by_class['func_instance_parms']:
             params_ent.remove()
-            for key, value in params_ent.keys.items():
+            for key, value in params_ent.items():
                 if not key.startswith('param'):
                     continue
                 # Don't bother parsing the index, it doesn't matter.
@@ -255,7 +283,7 @@ def get_inst_locs(map_filename: Path) -> FileSystemChain:
     """Given a map filename, find sdk_content and produce the lookup locations.
 
     The chained filesystem will first look relative to the map, then in
-    sdk_content/maps/ if that's a parent directory.
+    ``sdk_content/maps/`` if that's a parent directory.
     """
     fsys_rel = RawFileSystem(map_filename.parent)
     fsys = FileSystemChain(fsys_rel)
@@ -267,35 +295,45 @@ def get_inst_locs(map_filename: Path) -> FileSystemChain:
     return fsys
 
 
+def reset_keyvalue_warnings() -> None:
+    """If an unknown keyvalue is encountered during collapsing, a log message is produced only once.
+
+    This resets the internal tracker so the messages will be repeated again.
+    """
+    _UNKNOWN_KV.clear()
+
+
 def collapse_one(
     vmf: VMF,
     inst: Instance,
     file: InstanceFile,
-    fgd: FGD=None,
-    visgroup: Union[bool, VisGroup]=False,
+    *,
+    visgroup: Union[bool, VisGroup] = False,
+    engine_cache: MutableMapping[str, EntityDef] = srctools.EmptyMapping,
+    fgd: Optional[FGD] = None,
 ) -> None:
     """Collapse a single instance into the map.
 
-    The FGD is the data used to localise keyvalues. If none an internal database
-    will be used.
-    The visgroup paramter controls how visgroups are handled:
-    * If false, visgroups are stripped.
-    * If true, the original visgroups will be kept
-    * If set to a specific visgroup, all ents and brushes will be added to it,
-        with any existing visgroups in the instance added as a child.
+    The visgroup parameter controls how visgroups are handled:
+
+    * If :external:py:data:`False`, visgroups are stripped.
+    * If :external:py:data:`True`, the original visgroups will be kept.
+    * If set to a specific visgroup, all ents and brushes will be added to it, with any existing visgroups in the instance added as a child.
+
+    :param engine_cache: This is a dict used to cache parsed FGD data for remapping.
+        If multiple instances are being collapsed, create an empty dict and pass it in to every call
+        to allow reusing data.
+    :param fgd: :deprecated: old method of passing in a FGD database.
     """
     origin = inst.pos
     orient = inst.orient
     id_to_ent: Dict[int, Entity] = {}
 
-    if fgd is None:
-        fgd = FGD.engine_dbase()
-    # Contains all base-entity keyvalues, as a fallback.
-    try:
-        base_entity = fgd['_CBaseEntity_']
-    except KeyError:
-        LOGGER.warning('No CBaseEntity definition!')
-        base_entity = EntityDef(EntityTypes.BASE)
+    if fgd is not None:
+        warnings.warn(
+            'fgd parameter is deprecated, use engine_cache instead',
+            DeprecationWarning, stacklevel=2,
+        )
 
     if visgroup is not False:
         for old_group in file.vmf.vis_tree:
@@ -330,7 +368,7 @@ def collapse_one(
             if out.target.casefold() != folded_inst_name or out.inst_in is None:
                 continue
             try:
-                proxy_out = file.proxy_inputs[out.inst_in, out.input]  # type: ignore
+                proxy_out = file.proxy_inputs[out.inst_in, out.input]
             except KeyError:
                 # Not an error, could be another instance with our name.
                 continue
@@ -363,10 +401,10 @@ def collapse_one(
         id_to_ent[old_ent.id] = new_ent
 
         if visgroup is not False:
-            new_ent.visgroup_ids = [
-                inst.visgroup_ids[old] for old in
-                old_ent.visgroup_ids
-            ] or ungrouped_group.copy()
+            new_ent.visgroup_ids = {
+                inst.visgroup_ids[old]
+                for old in old_ent.visgroup_ids
+            } or ungrouped_group.copy()
 
         for old_brush, new_brush in zip(old_ent.solids, new_ent.solids):
             inst.brush_ids[old_brush.id] = new_brush.id
@@ -374,18 +412,27 @@ def collapse_one(
 
             # Convert across the IDs.
             if visgroup is not False:
-                new_brush.visgroup_ids = [
-                    inst.visgroup_ids[old] for old in
-                    old_brush.visgroup_ids
-                ] or ungrouped_group.copy()
+                new_brush.visgroup_ids = {
+                    inst.visgroup_ids[old]
+                    for old in old_brush.visgroup_ids
+                } or ungrouped_group.copy()
 
     for new_ent in new_ents:
         # Find the FGD to use.
         classname = new_ent['classname']
         try:
-            ent_type = fgd[classname]
+            ent_type = engine_cache[classname]
         except KeyError:
-            ent_type = base_entity
+            try:
+                ent_type = EntityDef.engine_def(classname)
+            except KeyError:
+                # Contains all base-entity keyvalues, as a fallback.
+                try:
+                    ent_type = EntityDef.engine_def('_CBaseEntity_')
+                except KeyError:
+                    LOGGER.warning('No CBaseEntity definition!')
+                    ent_type = EntityDef(EntityTypes.BASE)
+            engine_cache[classname] = ent_type
 
         # Set a hidden attribute to keep track of recursive instancing.
         if classname.casefold() == 'func_instance':
@@ -395,12 +442,12 @@ def collapse_one(
         # First extract a rotated angles value, handling the special "pitch" and "yaw" keys.
         angles = Angle.from_str(new_ent['angles'])
         if 'pitch' in new_ent:
-            angles.pitch = conv_float(new_ent['pitch'])
+            angles.pitch = srctools.conv_float(new_ent['pitch'])
         if 'yaw' in new_ent:
-            angles.yaw = conv_float(new_ent['yaw'])
+            angles.yaw = srctools.conv_float(new_ent['yaw'])
         angles @= orient
 
-        for key, value in new_ent.keys.items():
+        for key, value in new_ent.items():
             folded = key.casefold()
             value = inst.fixup.substitute(value, '')
             # Hardcode these critical keyvalues to always be these types.
@@ -410,11 +457,8 @@ def collapse_one(
             elif folded == 'angles':
                 new_ent['angles'] = str(angles)
                 continue
-            elif folded == 'pitch':
-                new_ent['pitch'] = str(angles.pitch)
-                continue
             elif folded == 'yaw':
-                new_ent['yaw'] = str(angles.yaw)
+                new_ent['yaw'] = format_float(angles.yaw)
                 continue
             elif folded in ('classname', 'hammerid', 'spawnflags', 'nodeid'):
                 continue
@@ -431,17 +475,28 @@ def collapse_one(
                 continue
             # This has specific interactions with angles, it needs to be the pitch KV.
             if kv.type is ValueTypes.ANGLE_NEG_PITCH:
-                if (classname, key) not in _UNKNOWN_KV:
+                if folded == 'pitch':
+                    new_ent['pitch'] = format_float(-angles.pitch)
+                elif (classname, key) not in _UNKNOWN_KV:
                     LOGGER.warning('angle_negative_pitch should only be applied to pitch, not {}.{}', classname, key)
                     _UNKNOWN_KV.add((classname, key))
                 continue
+            # Same, but not inverted.
+            elif kv.type is ValueTypes.EXT_ANGLE_PITCH:
+                if folded == 'pitch':
+                    new_ent['pitch'] = format_float(angles.pitch)
+                elif (classname, key) not in _UNKNOWN_KV:
+                    LOGGER.warning('angle_pitch should only be applied to pitch, not {}.{}', classname, key)
+                    _UNKNOWN_KV.add((classname, key))
+                continue
+            # Instance fixup commands shouldn't exist here.
             elif kv.type is ValueTypes.INST_VAR_REP:
                 if (classname, key) not in _UNKNOWN_KV:
                     LOGGER.warning('instance_variable should only be applied to replaceXX, not {}.{}', classname, key)
                     _UNKNOWN_KV.add((classname, key))
                 continue
 
-            new_ent.keys[key] = inst.fixup_key(vmf, fgd, kv.type, value)
+            new_ent[key] = inst.fixup_key(vmf, EntityDef.engine_classes(), kv.type, value)
 
         # Remap fixups on instance entities too.
         for key, value in new_ent.fixup.items():
@@ -467,26 +522,24 @@ def collapse_one(
 
 def collapse_all(
     vmf: VMF,
-    fsys: FileSystem,
-    recur_limit=100,
-    fgd: FGD=None,
+    fsys: FileSystem[Any],
+    recur_limit: int = 100,
 ) -> None:
-    """Searches for `func_instance`s in the map, then collapses them.
+    """Searches for ``func_instance`` in the map, then collapses them.
 
-    The filesystem is used to find the relevant instances.
-    The recursion limit indicates how many instances can be contained
-    in another - if it's exceeded they're left in the map.
+    :param vmf: The map to analyse.
+    :param fsys: The filesystem is used to find the relevant instances.
+    :param recur_limit: The recursion limit indicates how many instances can be contained in
+        another - if it's exceeded they're left in the map.
     """
-    if fgd is None:
-        fgd = FGD.engine_dbase()
-
     auto_inst_count = 0
+    fgd_cache: MutableMapping[str, EntityDef] = {}
+    file_cache: Dict[str, InstanceFile] = {}
 
-    cache: Dict[str, InstanceFile] = {}
     for _ in range(recur_limit):
         instances = list(vmf.by_class['func_instance'])
         if not instances:
-            break
+            return  # No more instances, success!
         for inst_ent in instances:
             inst = Instance.from_entity(inst_ent)
             inst_ent.remove()
@@ -495,11 +548,11 @@ def collapse_all(
                 auto_inst_count += 1
                 inst.name = f'InstanceAuto{auto_inst_count}'
             try:
-                file = cache[inst.filename]
+                file = file_cache[inst.filename]
             except KeyError:
-                props = fsys.read_prop(inst.filename)
+                props = fsys.read_kv1(inst.filename)
                 # except FileNotFoundError - fail.
-                file = cache[inst.filename] = InstanceFile(VMF.parse(props, preserve_ids=True))
-            collapse_one(vmf, inst, file, fgd)
-    else:  # Exhausted the range
+                file = file_cache[inst.filename] = InstanceFile(VMF.parse(props, preserve_ids=True))
+            collapse_one(vmf, inst, file, engine_cache=fgd_cache)
+    else:  # Exhausted the range, we must have too much recursion in the instances.
         raise RecursionError('Loop in instances!')
