@@ -1,17 +1,16 @@
 """Parses SMD model/animation data."""
-import os
-import re
-import warnings
+from typing import Any, Dict, Iterable, Iterator, List, Optional, Set, Tuple, Union
+from typing_extensions import Protocol
 from collections import defaultdict
 from copy import deepcopy
 from operator import itemgetter
-
 import math
-from typing import (
-    Union, Optional, Any, Protocol, Tuple,
-    List, Dict, Set, Iterator, Iterable,
-)
-from srctools.math import Vec, Angle, Matrix, to_matrix
+import os
+import re
+import warnings
+
+from srctools.math import Angle, Matrix, Vec, to_matrix
+
 
 __all__ = [
     'Mesh', 'Triangle', 'Vertex', 'Bone', 'BoneFrame', 'ParseError',
@@ -43,7 +42,7 @@ class Bone:
     def __copy__(self) -> 'Bone':
         return Bone(self.name, self.parent)
 
-    def __deepcopy__(self, memodict: dict=None) -> 'Bone':
+    def __deepcopy__(self, memodict: Optional[Dict[int, Any]] = None) -> 'Bone':
         return Bone(self.name, deepcopy(self.parent, memodict))
 
     def __eq__(self, other: Any) -> bool:
@@ -80,7 +79,7 @@ class BoneFrame:
     def __copy__(self) -> 'BoneFrame':
         return BoneFrame(self.bone, self.position, self.rotation)
 
-    def __deepcopy__(self, memodict: dict=None) -> 'BoneFrame':
+    def __deepcopy__(self, memodict: Optional[Dict[int, Any]] = None) -> 'BoneFrame':
         return BoneFrame(
             deepcopy(self.bone, memodict),
             self.position.copy(),
@@ -106,8 +105,10 @@ class Vertex:
         self.tex_v = tex_v
 
     def __repr__(self) -> str:
-        return 'Vertex({!r}, {!r}, {}, {}, {})'.format(
-            self.pos, self.norm, self.tex_u, self.tex_v, self.links
+        return (
+            f'{self.__class__.__name__}('
+            f'{self.pos!r}, {self.norm!r}, '
+            f'{self.tex_u}, {self.tex_v}, {self.links})'
         )
 
     def copy(self) -> 'Vertex':
@@ -122,7 +123,7 @@ class Vertex:
 
     __copy__ = copy
 
-    def __deepcopy__(self, memodict: dict=None) -> 'Vertex':
+    def __deepcopy__(self, memodict: Optional[Dict[int, Any]] = None) -> 'Vertex':
         return Vertex(
             self.pos.copy(),
             self.norm.copy(),
@@ -131,7 +132,7 @@ class Vertex:
             deepcopy(self.links, memodict),
         )
 
-    def with_pos(self, pos: Vec, norm: Vec=None) -> 'Vertex':
+    def with_pos(self, pos: Vec, norm: Optional[Vec] = None) -> 'Vertex':
         """Copy this vertex, changing the position."""
         if norm is None:
             norm = self.norm
@@ -195,7 +196,7 @@ class Triangle:
 
     __copy__ = copy
 
-    def __deepcopy__(self, memodict: dict=None) -> 'Triangle':
+    def __deepcopy__(self, memodict: Optional[Dict[int, Any]] = None) -> 'Triangle':
         """Duplicate this triangle."""
         return Triangle(
             self.mat,
@@ -269,13 +270,21 @@ class Mesh:
     def __copy__(self) -> 'Mesh':
         return Mesh(self.bones, self.animation, self.triangles)
 
-    def __deepcopy__(self, memodict: dict=None) -> 'Mesh':
+    def __deepcopy__(self, memodict: Optional[Dict[int, Any]] = None) -> 'Mesh':
         return Mesh(
             deepcopy(self.bones, memodict),
             deepcopy(self.animation, memodict),
             deepcopy(self.triangles, memodict),
         )
 
+    def root_bone(self) -> Bone:
+        """Return the root of our bone hierachy."""
+        for bone in self.bones.values():
+            if bone.parent is None:
+                return bone
+        else:
+            raise ValueError('No root bone?')
+
     @staticmethod
     def blank(root_name: str) -> 'Mesh':
         """Create an empty mesh, with a single root bone."""
@@ -359,14 +368,18 @@ class Mesh:
         for line_num, line in file_iter:
             if line == b'end':
                 return bones
+            match = re.fullmatch(
+                br'([0-9]+)\s*"([^"]*)"\s*(-?[0-9]+)',
+                line,
+            )
+            if match is None:
+                raise ParseError(line_num, 'Invalid line!') from None
+
+            bone_ind_bytes, bone_name, bone_parent_bytes = match.groups()
             try:
-                bone_ind_bytes, bone_name, bone_parent_bytes = re.fullmatch(
-                    br'([0-9]+)\s*"([^"]*)"\s*(-?[0-9]+)',
-                    line,
-                ).groups()
                 bone_ind = int(bone_ind_bytes)
                 bone_parent = int(bone_parent_bytes)
-            except (ValueError, AttributeError):  # None.groups()
+            except ValueError:  # None.groups()
                 raise ParseError(line_num, 'Invalid line!') from None
             else:
                 if bone_parent == -1:
@@ -422,7 +435,7 @@ class Mesh:
         raise ParseError('end', 'No end to skeleton section!')
 
     @staticmethod
-    def _parse_smd_tri(file_iter: Iterator[Tuple[int, bytes]], bones: Dict[int, Bone]):
+    def _parse_smd_tri(file_iter: Iterator[Tuple[int, bytes]], bones: Dict[int, Bone]) -> List[Triangle]:
         """Parse the 'triangles' section of SMDs."""
         tris: List[Triangle] = []
         # Temporary vertex, which we overwrite in the loop.
@@ -568,8 +581,8 @@ class Mesh:
     def append_model(
         self,
         mdl: 'Mesh',
-        rotation: Union[Angle, Matrix, Vec, None]=None,
-        offset: Vec=None,
+        rotation: Union[Angle, Matrix, Vec, None] = None,
+        offset: Optional[Vec] = None,
         scale: float=1.0,
     ) -> None:
         """Append another model's geometry onto this one.
@@ -582,14 +595,7 @@ class Mesh:
         if offset is None:
             offset = Vec()
 
-        for bone in self.bones.values():
-            if bone.parent is None:
-                root_bone = bone
-                break
-        else:
-            raise ValueError('No root bone?')
-
-        bone_link = [(root_bone, 1.0)]
+        bone_link = [(self.root_bone(), 1.0)]
         matrix = to_matrix(rotation)
 
         for orig_tri in mdl.triangles:
