@@ -3,45 +3,47 @@
 After this is imported, the imghdr module can recoginise
 VTF images (returning 'source_vtf').
 """
+from typing import (
+    IO, TYPE_CHECKING, Any, Collection, Dict, Iterable, List, Mapping, Optional, Sequence,
+    Tuple, Type, Union, overload,
+)
 from array import array
-from enum import Enum, Flag
 from collections import namedtuple
+from enum import Enum, Flag
+from io import BytesIO
 import itertools
 import math
 import struct
+import types
 import warnings
-from io import BytesIO
 
-from srctools import binformat, EmptyMapping
-from srctools.const import add_unknown
-from srctools.math import Vec
+from . import EmptyMapping, binformat
+from .const import add_unknown
+from .math import Vec
 
-from typing import (
-    IO, Dict, List, Optional, Tuple, Iterable, Union,
-    TYPE_CHECKING, Type, Collection, overload, Sequence,
-    Mapping, BinaryIO,
-)
 
 # Only import while type checking, so these expensive libraries are only loaded
 # if the user used them elsewhere.
 if TYPE_CHECKING:
-    from PIL.Image import Image as PIL_Image
     import tkinter
-    import wx
+
+    from PIL.Image import Image as PIL_Image
 
 # A little dance to import both the Cython and Python versions,
 # and choose an appropriate unprefixed version.
 # For type-checking purposes make it think the Cython version is the Python one.
 
 # noinspection PyProtectedMember
-from srctools import _py_vtf_readwrite as _py_format_funcs
+from . import _py_vtf_readwrite as _py_format_funcs
+
+
 _cy_format_funcs = _format_funcs = _py_format_funcs
 
 if not TYPE_CHECKING:
     try:
         # noinspection PyUnresolvedReferences, PyProtectedMember
-        from srctools import _cy_vtf_readwrite as _cy_format_funcs  # type: ignore
-        _format_funcs = _cy_format_funcs  # type: ignore
+        from . import _cy_vtf_readwrite as _cy_format_funcs
+        _format_funcs = _cy_format_funcs
     except ImportError:
         pass
 
@@ -80,13 +82,13 @@ _BLANK_PIXEL = array('B', [0, 0, 0, 0xFF])
 def _mk_fmt(
     r: int = 0, g: int = 0, b: int = 0,
     a: int = 0, *,
-    l: int = 0, size: int = 0,
+    grey: int = 0, size: int = 0,
 ) -> Tuple[int, int, int, int, int, int]:
     """Helper function to construct ImageFormats."""
     global _mk_fmt_ind
-    if l:
-        r = g = b = l
-        size = l + a
+    if grey:
+        r = g = b = grey
+        size = grey + a
     if not size:
         size = r + g + b + a
     _mk_fmt_ind += 1
@@ -99,7 +101,7 @@ _mk_fmt_ind = -1  # Incremented first time to 0
 
 class ImageFormats(Enum):
     """All VTF image formats, with their data sizes in the value."""
-    def __init__(self, r, g, b, a, size, ind) -> None:
+    def __init__(self, r: int, g: int, b: int, a: int, size: int, ind: int) -> None:
         self.r = r
         self.g = g
         self.b = b
@@ -112,8 +114,8 @@ class ImageFormats(Enum):
     RGB888 = _mk_fmt(8, 8, 8, 0)
     BGR888 = _mk_fmt(8, 8, 8)
     RGB565 = _mk_fmt(5, 6, 5, 0)
-    I8 = _mk_fmt(a=0, l=8)
-    IA88 = _mk_fmt(a=8, l=8)
+    I8 = _mk_fmt(a=0, grey=8)
+    IA88 = _mk_fmt(a=8, grey=8)
     P8 = _mk_fmt()  # Using a palette somehow - was never implemented by Valve.
     A8 = _mk_fmt(a=8)
     # Blue = alpha channel too
@@ -303,13 +305,12 @@ class Frame:
 
     This is lazy, so it will only read from the file when actually used.
     """
+    __slots__ = ['width', 'height', '_data', '_fileinfo']
+    width: int
+    height: int
+    _data: Optional['array[int]']  # Only generic in stubs!
+    _fileinfo: Optional[Tuple[IO[bytes], int, ImageFormats]]
 
-    __slots__ = [
-        'width',
-        'height',
-        '_data',
-        '_fileinfo',
-    ]
     def __init__(
         self,
         width: int,
@@ -318,8 +319,8 @@ class Frame:
         """Private constructor, creates a blank image of this size."""
         self.width = width
         self.height = height
-        self._data = None  # type: Optional[array]
-        self._fileinfo = None  # type: Optional[Tuple[IO[bytes], int, ImageFormats]]
+        self._data = None
+        self._fileinfo = None
 
     def load(self) -> None:
         """If the image has not been loaded, load it from the file stream."""
@@ -357,13 +358,13 @@ class Frame:
     @overload
     def copy_from(
         self,
-        source: Union[bytes, bytearray, array, memoryview],
+        source: Union[bytes, bytearray, 'array[int]', memoryview],
         format: ImageFormats = ImageFormats.RGBA8888,
     ) -> None: ...
 
     def copy_from(
         self,
-        source: Union['Frame', bytes, bytearray, array, memoryview],
+        source: Union['Frame', bytes, bytearray, 'array[int]', memoryview],
         format: ImageFormats = ImageFormats.RGBA8888,
     ) -> None:
         """Overwrite this frame with other data.
@@ -467,7 +468,12 @@ class Frame:
             1,
         ).copy()
 
-    def to_tkinter(self, tk: 'tkinter.Misc' = None, *, bg: Optional[Tuple[int, int, int]]=None) -> 'tkinter.PhotoImage':
+    def to_tkinter(
+        self,
+        tk: 'tkinter.Misc | None' = None,
+        *,
+        bg: Optional[Tuple[int, int, int]] = None,
+    ) -> 'tkinter.PhotoImage':
         """Convert the given frame into a Tkinter PhotoImage.
 
         If bg is set, the image will be composited onto this background.
@@ -490,8 +496,9 @@ class Frame:
             ),
         )
 
-    def to_wx_image(self, bg: Optional[Tuple[int, int, int]]=None) -> 'wx.Image':
-        """Convert the given frame into a wxPython image.
+    # TODO: wx has no type hints, so we can't import.
+    def to_wx_image(self, bg: Optional[Tuple[int, int, int]]=None) -> Any:
+        """Convert the given frame into a wxPython wx.Image.
 
         This requires wxPython to be installed.
         If bg is set, the image will be composited onto this background.
@@ -499,14 +506,14 @@ class Frame:
         """
         self.load()
         assert self._data is not None
-        import wx
+        import wx  # type: ignore
 
         img = wx.Image(self.width, self.height)
         _format_funcs.alpha_flatten(self._data, img.GetDataBuffer(), self.width, self.height, bg)
         return img
 
-    def to_wx_bitmap(self, bg: Optional[Tuple[int, int, int]]=None) -> 'wx.Bitmap':
-        """Convert the given frame into a wxPython bitmap.
+    def to_wx_bitmap(self, bg: Optional[Tuple[int, int, int]]=None) -> Any:
+        """Convert the given frame into a wxPython wx.Bitmap.
 
         This requires wxPython to be installed.
         If bg is set, the image will be composited onto this background.
@@ -514,7 +521,7 @@ class Frame:
         """
         self.load()
         assert self._data is not None
-        import wx
+        import wx  # pyright: ignore
 
         img = wx.Bitmap(self.width, self.height)
         # Bitmap memory layout isn't public, so we have to write to a temporary
@@ -527,6 +534,24 @@ class Frame:
 
 class VTF:
     """Valve Texture Format files, used in the Source Engine."""
+    width: int
+    height: int
+    depth: int
+    mipmap_count: int
+
+    version: Tuple[int, int]
+    reflectivity: Vec
+    bumpmap_scale: float
+    resources: Dict[Union[ResourceID, bytes], Resource]
+    sheet_info: Dict[int, 'SheetSequence']
+    flags: VTFFlags
+    frame_count: int
+    format: ImageFormats
+    low_format: ImageFormats
+
+    _frames: Dict[Tuple[int, Union[CubeSide, int], int], Frame]
+    _low_res: Frame
+
     def __init__(
         self,
         width: int,
@@ -568,7 +593,7 @@ class VTF:
         self.version = version
         self.reflectivity = ref
         self.bumpmap_scale = bump_scale
-        self.resources = {}  # type: Dict[Union[ResourceID, bytes], Resource]
+        self.resources = {}
         self.sheet_info = dict(sheet_info)
         self.flags = flags
         self.frame_count = frames
@@ -576,12 +601,13 @@ class VTF:
         self.low_format = thumb_fmt
 
         # (frame, depth/cubemap, mipmap) -> frame
-        self._frames = {}  # type: Dict[Tuple[int, Union[CubeSide, int], int], Frame]
+        self._frames = {}
         self._low_res = Frame(16, 16)
 
+        depth_iter: Iterable[Union[int, CubeSide]]
         if VTFFlags.ENVMAP in flags:
             if version[1] == 5:
-                depth_iter = CUBES  # type: Iterable[Union[int, CubeSide]]
+                depth_iter = CUBES
             else:
                 depth_iter = CUBES_WITH_SPHERE
         else:
@@ -620,6 +646,9 @@ class VTF:
 
         vtf = cls.__new__(cls)
 
+        header_size: int
+        low_width: int
+        low_height: int
         (
             header_size,
             width,
@@ -658,8 +687,8 @@ class VTF:
         if vtf.depth <= 0:
             vtf.depth = 1
 
-        low_res_offset: Optional[int] = None
-        high_res_offset: Optional[int] = None
+        low_res_offset = -1
+        high_res_offset = -1
 
         vtf.resources = {}
         vtf.sheet_info = {}
@@ -668,7 +697,10 @@ class VTF:
         if version_minor >= 3:
             [num_resources] = struct.unpack('<3xI8x', file.read(15))
             for i in range(num_resources):
-                [res_id, res_flags, data] = struct.unpack('<3sBI', file.read(8))  # type: bytes, int, int
+                res_id: bytes
+                res_flags: int
+                data: int
+                [res_id, res_flags, data] = struct.unpack('<3sBI', file.read(8))
                 if res_id in vtf.resources:
                     raise ValueError(f'Duplicate resource ID {repr(res_id)[1:]}!')
 
@@ -701,7 +733,7 @@ class VTF:
             low_res_offset = header_size
             high_res_offset = low_res_offset + low_fmt.frame_size(low_width, low_height)
 
-        if high_res_offset is None:
+        if high_res_offset < 0:
             raise ValueError('Missing main image resource!')
 
         # We don't implement these high-res formats.
@@ -710,16 +742,17 @@ class VTF:
 
         vtf._low_res = Frame(low_width, low_height)
         if low_fmt is not ImageFormats.NONE:
-            if low_res_offset is None:
+            if low_res_offset < 0:
                 raise ValueError('Missing low-res thumbnail resource!')
             vtf._low_res._fileinfo = (file, low_res_offset, low_fmt)
 
         # If cubemaps are present, we iterate that for depth.
         # Otherwise it's the depth value.
+        depth_iter: Iterable[Union[int, CubeSide]]
         if VTFFlags.ENVMAP in vtf.flags:
             # For version 7.5, the spheremap is skipped.
             if version_minor == 5:
-                depth_iter = CUBES  # type: Iterable[Union[int, CubeSide]]
+                depth_iter = CUBES
             else:
                 depth_iter = CUBES_WITH_SPHERE
         else:
@@ -879,7 +912,10 @@ class VTF:
         """
         return self
 
-    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
+    def __exit__(
+        self,
+        exc_type: Type[BaseException], exc_val: BaseException, exc_tb: types.TracebackType,
+    ) -> None:
         """Close the streams if any frames still have them open."""
         for frame in self._frames.values():
             frame._fileinfo = None
@@ -946,7 +982,7 @@ class VTF:
         self, *,
         frame: int = 0,
         depth: int = 0,
-        side: CubeSide = None,
+        side: Optional[CubeSide] = None,
         mipmap: int = 0,
     ) -> Frame:
         """Get a specific image frame.
@@ -983,7 +1019,7 @@ class SheetSequence:
         self.duration = duration
 
     @classmethod
-    def from_resource(cls, data: bytes) -> Dict[str, 'SheetSequence']:
+    def from_resource(cls, data: bytes) -> Dict[int, 'SheetSequence']:
         """Decode from the resource data."""
         (
             version,
@@ -994,7 +1030,7 @@ class SheetSequence:
         if version > 1:
             raise ValueError('Unknown version {}!'.format(version))
 
-        sequences: Dict[str, SheetSequence] = {}
+        sequences: Dict[int, SheetSequence] = {}
         if sequence_count > SheetSequence.MAX_COUNT:
             raise ValueError('Cannot have more than {} sequences ({})!'.format(
                 SheetSequence.MAX_COUNT,
@@ -1014,7 +1050,7 @@ class SheetSequence:
             if seq_num in sequences:
                 raise ValueError('Duplicate sequence number {}!'.format(seq_num))
 
-            frames = []  # type: List[Tuple[float, TexCoord, TexCoord, TexCoord, TexCoord]]
+            frames: List[Tuple[float, TexCoord, TexCoord, TexCoord, TexCoord]] = []
 
             for frame_ind in range(frame_count):
                 [duration] = struct.unpack_from('<f', data, offset)
@@ -1064,22 +1100,3 @@ class SheetSequence:
                     file.write(struct.pack('<4f', *tex_d))
 
         return file.getvalue()
-
-# Add support for the imghdr module.
-
-
-def test_vtf(h: bytes, f: Optional[BinaryIO]) -> Optional[str]:
-    """Source Engine Valve Texture Format."""
-    if h[:4] == b'VTF\0':
-        try:
-            version_major, version_minor = struct.unpack('II', h[4:12])
-        except struct.error:
-            return None
-        if version_major == 7 and (0 <= version_minor <= 5):
-            return 'source_vtf'
-    return None
-
-import imghdr
-imghdr.test_vtf = test_vtf  # type: ignore
-imghdr.tests.append(test_vtf)
-del imghdr, test_vtf
